package ru.dbotthepony.kstarbound.world

import ru.dbotthepony.kbox2d.api.BodyDef
import ru.dbotthepony.kbox2d.api.FixtureDef
import ru.dbotthepony.kbox2d.collision.shapes.PolygonShape
import ru.dbotthepony.kbox2d.dynamics.B2Fixture
import ru.dbotthepony.kstarbound.defs.TileDefinition
import ru.dbotthepony.kstarbound.math.*
import ru.dbotthepony.kstarbound.world.entities.Entity
import ru.dbotthepony.kvector.api.IStruct2d
import ru.dbotthepony.kvector.api.IStruct2i
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.nint.Vector2i
import java.util.*
import kotlin.collections.ArrayList
import kotlin.collections.HashSet
import kotlin.math.absoluteValue

/**
 * Представляет из себя класс, который содержит состояние тайла на заданной позиции
 */
data class ChunkTile(val chunk: Chunk<*, *>.TileLayer, val def: TileDefinition) {
	var color = 0
		set(value) {
			field = value
			chunk.incChangeset()
		}

	var forceVariant = -1
		set(value) {
			field = value
			chunk.incChangeset()
		}
}

interface ITileMap {
	/**
	 * Относительная проверка находится ли координата вне границ чанка
	 */
	fun isOutside(x: Int, y: Int): Boolean {
		return x !in 0 until CHUNK_SIZE || y !in 0 until CHUNK_SIZE
	}
}

/**
 * Предоставляет интерфейс для доступа к тайлам в чанке
 */
interface ITileGetter : ITileMap {
	/**
	 * Возвращает тайл по ОТНОСИТЕЛЬНЫМ координатам внутри чанка
	 */
	operator fun get(x: Int, y: Int): ChunkTile?

	/**
	 * Возвращает тайл по ОТНОСИТЕЛЬНЫМ координатам внутри чанка
	 */
	operator fun get(pos: Vector2i) = get(pos.x, pos.y)

	/**
	 * Возвращает итератор пар <Vector2i, Тайл?>
	 *
	 * Вектор имеет ОТНОСИТЕЛЬНЫЕ значения внутри самого чанка
	 */
	val posToTile: Iterator<Pair<Vector2i, ChunkTile?>> get() {
		return object : Iterator<Pair<Vector2i, ChunkTile?>> {
			private var x = 0
			private var y = 0

			private fun idx() = x + CHUNK_SIZE * y

			override fun hasNext(): Boolean {
				return idx() < CHUNK_SIZE * CHUNK_SIZE
			}

			override fun next(): Pair<Vector2i, ChunkTile?> {
				if (!hasNext()) {
					throw IllegalStateException("Already iterated everything!")
				}

				val tile = this@ITileGetter[x, y]
				val pos = Vector2i(x, y)

				x++

				if (x >= CHUNK_SIZE) {
					y++
					x = 0
				}

				return pos to tile
			}
		}
	}
}

/**
 * Интерфейс предоставляет из себя описание класса, который имеет координаты чанка
 */
interface IChunkPositionable : ITileMap {
	val pos: ChunkPos

	/**
	 * Возвращает псевдослучайное Long для заданной позиции
	 *
	 * Для использования в рендерах и прочих вещах, которым нужно стабильное число на основе своей позиции
	 */
	fun randomLongFor(x: Int, y: Int): Long {
		var long = (x or (pos.x shl CHUNK_SHIFT)) * 738548L + (y or (pos.y shl CHUNK_SHIFT)) * 2191293543L
		long = long xor 8339437585692L
		long = (long ushr 4) or (long shl 52)
		long *= 7848344324L
		long = (long ushr 12) or (long shl 44)
		return long
	}

	/**
	 * Возвращает псевдослучайное нормализированное Double для заданной позиции
	 *
	 * Для использования в рендерах и прочих вещах, которым нужно стабильное число на основе своей позиции
	 */
	fun randomDoubleFor(x: Int, y: Int): Double {
		return (randomLongFor(x, y) / 9.223372036854776E18) / 2.0 + 0.5
	}

	/**
	 * Возвращает псевдослучайное Long для заданной позиции
	 *
	 * Для использования в рендерах и прочих вещах, которым нужно стабильное число на основе своей позиции
	 */
	fun randomLongFor(pos: Vector2i) = randomLongFor(pos.x, pos.y)

	/**
	 * Возвращает псевдослучайное нормализированное Double для заданной позиции
	 *
	 * Для использования в рендерах и прочих вещах, которым нужно стабильное число на основе своей позиции
	 */
	fun randomDoubleFor(pos: Vector2i) = randomDoubleFor(pos.x, pos.y)
}

/**
 * Предоставляет интерфейс по установке тайлов в чанке
 */
interface ITileSetter : ITileMap {
	/**
	 * Устанавливает тайл по ОТНОСИТЕЛЬНЫМ координатам внутри чанка
	 */
	operator fun set(x: Int, y: Int, tile: TileDefinition?): ChunkTile?
	/**
	 * Устанавливает тайл по ОТНОСИТЕЛЬНЫМ координатам внутри чанка
	 */
	operator fun set(pos: Vector2i, tile: TileDefinition?) = set(pos.x, pos.y, tile)
}

interface ITileGetterSetter : ITileGetter, ITileSetter
interface ITileChunk : ITileGetter, IChunkPositionable
interface IMutableTileChunk : ITileChunk, ITileSetter

const val CHUNK_SHIFT = 5
const val CHUNK_SIZE = 1 shl CHUNK_SHIFT // 32

const val CHUNK_SIZE_FF = CHUNK_SIZE - 1
const val CHUNK_SIZEf = CHUNK_SIZE.toFloat()
const val CHUNK_SIZEd = CHUNK_SIZE.toDouble()

class ChunkPos(val x: Int, val y: Int) : Comparable<ChunkPos> {
	constructor(pos: IStruct2i) : this(pos.component1(), pos.component2())

	val firstBlock get() = Vector2i(x shl CHUNK_SHIFT, y shl CHUNK_SHIFT)
	val lastBlock get() = Vector2i(((x + 1) shl CHUNK_SHIFT) - 1, ((y + 1) shl CHUNK_SHIFT) - 1)

	fun up(): ChunkPos {
		return ChunkPos(x, y + 1)
	}

	fun down(): ChunkPos {
		return ChunkPos(x, y - 1)
	}

	fun left(): ChunkPos {
		return ChunkPos(x + 1, y)
	}

	fun right(): ChunkPos {
		return ChunkPos(x - 1, y)
	}

	override fun equals(other: Any?): Boolean {
		if (other is ChunkPos)
			return other.x == x && other.y == y

		return false
	}

	override fun hashCode(): Int {
		return (y shl 16) xor x
	}

	override fun toString(): String {
		return "ChunkPos[$x $y]"
	}

	override fun compareTo(other: ChunkPos): Int {
		if (x > other.x) {
			return 1
		} else if (x < other.x) {
			return -1
		}

		return y.compareTo(other.y)
	}

	companion object {
		val ZERO = ChunkPos(0, 0)

		fun fromTilePosition(input: IStruct2i): ChunkPos {
			val (x, y) = input
			return ChunkPos(tileToChunkComponent(x), tileToChunkComponent(y))
		}

		fun fromTilePosition(input: IStruct2d): ChunkPos {
			val (x, y) = input
			return fromTilePosition(x, y)
		}

		fun fromTilePosition(x: Int, y: Int): ChunkPos {
			return ChunkPos(tileToChunkComponent(x), tileToChunkComponent(y))
		}

		fun fromTilePosition(x: Double, y: Double): ChunkPos {
			return ChunkPos(tileToChunkComponent(roundByAbsoluteValue(x)), tileToChunkComponent(roundByAbsoluteValue(y)))
		}

		fun normalizeCoordinate(input: Int): Int {
			val band = input and CHUNK_SIZE_FF

			if (band < 0) {
				return band + CHUNK_SIZE_FF
			}

			return band
		}

		fun tileToChunkComponent(comp: Int): Int {
			if (comp < 0) {
				return -(comp.absoluteValue shr CHUNK_SHIFT) - 1
			}

			return comp shr CHUNK_SHIFT
		}
	}
}

/**
 * Предоставляет доступ к чанку и его соседям
 *
 * В основном для использования в местах, где нужен не мир, а определённый чанк мира,
 * и при этом координаты проверяются относительно чанка и могут спокойно выйти за его пределы,
 * с желанием получить тайл из соседнего чанка
 */
open class TileChunkView(
	open val center: ITileChunk,

	open val right: ITileChunk?,
	open val top: ITileChunk?,
	open val topRight: ITileChunk?,
	open val topLeft: ITileChunk?,

	open val left: ITileChunk?,
	open val bottom: ITileChunk?,
	open val bottomLeft: ITileChunk?,
	open val bottomRight: ITileChunk?,
) : ITileChunk {
	override fun get(x: Int, y: Int): ChunkTile? {
		if (x in 0 .. CHUNK_SIZE_FF) {
			if (y in 0 .. CHUNK_SIZE_FF) {
				return center[x, y]
			}

			if (y < 0) {
				return bottom?.get(x, y + CHUNK_SIZE)
			} else {
				return top?.get(x, y - CHUNK_SIZE)
			}
		}

		if (x < 0) {
			if (y in 0 .. CHUNK_SIZE_FF) {
				return left?.get(x + CHUNK_SIZE, y)
			}

			if (y < 0) {
				return bottomLeft?.get(x + CHUNK_SIZE, y + CHUNK_SIZE)
			} else {
				return topLeft?.get(x + CHUNK_SIZE, y - CHUNK_SIZE)
			}
		} else {
			if (y in 0 .. CHUNK_SIZE_FF) {
				return right?.get(x - CHUNK_SIZE, y)
			}

			if (y < 0) {
				return bottomRight?.get(x - CHUNK_SIZE, y + CHUNK_SIZE)
			} else {
				return topRight?.get(x - CHUNK_SIZE, y - CHUNK_SIZE)
			}
		}
	}

	override val pos: ChunkPos
		get() = center.pos
}

class MutableTileChunkView(
	override val center: IMutableTileChunk,

	override val right: IMutableTileChunk?,
	override val top: IMutableTileChunk?,
	override val topRight: IMutableTileChunk?,
	override val topLeft: IMutableTileChunk?,

	override val left: IMutableTileChunk?,
	override val bottom: IMutableTileChunk?,
	override val bottomLeft: IMutableTileChunk?,
	override val bottomRight: IMutableTileChunk?,
) : TileChunkView(center, right, top, topRight, topLeft, left, bottom, bottomLeft, bottomRight), IMutableTileChunk {
	override fun set(x: Int, y: Int, tile: TileDefinition?): ChunkTile? {
		if (x in 0 .. CHUNK_SIZE_FF) {
			if (y in 0 .. CHUNK_SIZE_FF) {
				return center.set(x, y, tile)
			}

			if (y < 0) {
				return bottom?.set(x, y + CHUNK_SIZE, tile)
			} else {
				return top?.set(x, y - CHUNK_SIZE, tile)
			}
		}

		if (x < 0) {
			if (y in 0 .. CHUNK_SIZE_FF) {
				return left?.set(x + CHUNK_SIZE, y, tile)
			}

			if (y < 0) {
				return bottomLeft?.set(x + CHUNK_SIZE, y + CHUNK_SIZE, tile)
			} else {
				return topLeft?.set(x + CHUNK_SIZE, y - CHUNK_SIZE, tile)
			}
		} else {
			if (y in 0 .. CHUNK_SIZE_FF) {
				return right?.set(x - CHUNK_SIZE, y, tile)
			}

			if (y < 0) {
				return bottomRight?.set(x - CHUNK_SIZE, y + CHUNK_SIZE, tile)
			} else {
				return topRight?.set(x - CHUNK_SIZE, y - CHUNK_SIZE, tile)
			}
		}
	}
}

private data class TileExposure(
	var pos: Vector2d,
	val left: Boolean,
	val right: Boolean,
	val up: Boolean,
	val down: Boolean,
)

private class TileFlooder(
	private val tiles: Array<ChunkTile?>,
	private val seen: BooleanArray,
	rootx: Int,
	rooty: Int
) {
	/**
	 * Tile positions which have at least one face free of neighbours
	 */
	val exposed = ArrayList<TileExposure>()

	private fun get(x: Int, y: Int): Boolean {
		if (x < 0 || x > CHUNK_SIZE_FF) {
			return false
		}

		if (y < 0 || y > CHUNK_SIZE_FF) {
			return false
		}

		return tiles[x or (y shl CHUNK_SHIFT)] != null
	}

	private fun visit(x: Int, y: Int) {
		if (seen[x or (y shl CHUNK_SHIFT)])
			return

		seen[x or (y shl CHUNK_SHIFT)] = true

		val left = get(x - 1, y)
		val right = get(x + 1, y)
		val up = get(x, y + 1)
		val down = get(x, y - 1)

		if (!left || !right || !up || !down) {
			exposed.add(TileExposure(Vector2d(x.toDouble() + 0.5, y.toDouble() + 0.5), !left, !right, !up, !down))
		}

		if (left) {
			visit(x - 1, y)
		}

		if (right) {
			visit(x + 1, y)
		}

		if (up) {
			visit(x, y + 1)
		}

		if (down) {
			visit(x, y - 1)
		}
	}

	init {
		visit(rootx, rooty)
	}
}

private class RectTileFlooderDepthFirst(
	private val tiles: Array<ChunkTile?>,
	private val seen: BooleanArray,
	rootx: Int,
	rooty: Int
) {
	val mins: Vector2i
	val maxs: Vector2i

	private fun filled(x: Int, y: Int): Boolean {
		if (x < 0 || x > CHUNK_SIZE_FF) {
			return false
		}

		if (y < 0 || y > CHUNK_SIZE_FF) {
			return false
		}

		return !seen[x or (y shl CHUNK_SHIFT)] && tiles[x or (y shl CHUNK_SHIFT)] != null
	}

	init {
		// expand wide
		var widthPositive = 1

		while (true) {
			if (!filled(rootx + widthPositive, rooty)) {
				break
			}

			widthPositive++
		}

		var widthNegative = 1

		while (true) {
			if (!filled(rootx - widthNegative, rooty)) {
				break
			}

			widthNegative++
		}

		// expand tall
		var heightPositive = 1

		while (true) {
			if (!filled(rootx, rooty + heightPositive)) {
				break
			}

			heightPositive++
		}

		var heightNegative = 1

		while (true) {
			if (!filled(rootx, rooty - heightNegative)) {
				break
			}

			heightNegative++
		}

		widthPositive -= 1
		widthNegative -= 1

		heightNegative -= 1
		heightPositive -= 1

		if (heightPositive + heightNegative > widthPositive + widthNegative) {
			// height is bigger
			// try to expand wide
			widthPositive = 0
			widthNegative = 0

			while (true) {
				var escape = false

				for (i in -heightNegative .. heightPositive) {
					if (!filled(rootx + widthPositive, rooty + i)) {
						escape = true
						break
					}
				}

				if (escape) {
					break
				}

				widthPositive++
			}

			while (true) {
				var escape = false

				for (i in -heightNegative .. heightPositive) {
					if (!filled(rootx - widthNegative, rooty + i)) {
						escape = true
						break
					}
				}

				if (escape) {
					break
				}

				widthNegative++
			}

			widthNegative -= 1
			widthPositive -= 1
		} else {
			// height is equal or lesser than width
			// expand high
			heightNegative = 0
			heightPositive = 0

			while (true) {
				var escape = false

				for (i in -widthNegative .. widthPositive) {
					if (!filled(rootx + i, rooty + heightPositive)) {
						escape = true
						break
					}
				}

				if (escape) {
					break
				}

				heightPositive++
			}

			while (true) {
				var escape = false

				for (i in -widthNegative .. widthPositive) {
					if (!filled(rootx + i, rooty - heightNegative)) {
						escape = true
						break
					}
				}

				if (escape) {
					break
				}

				heightNegative++
			}

			heightNegative -= 1
			heightPositive -= 1
		}

		mins = Vector2i(rootx - widthNegative, rooty - heightNegative)
		maxs = Vector2i(rootx + widthPositive, rooty + heightPositive)
	}

	fun markSeen() {
		for (x in mins.x .. maxs.x) {
			for (y in mins.y .. maxs.y) {
				seen[x or (y shl CHUNK_SHIFT)] = true
			}
		}
	}
}

private class RectTileFlooderSizeFirst(
	private val tiles: Array<ChunkTile?>,
	private val seen: BooleanArray,
	private val rootx: Int,
	private val rooty: Int
) {
	val mins: Vector2i
	val maxs: Vector2i

	private fun filled(x: Int, y: Int): Boolean {
		if (x < 0 || x > CHUNK_SIZE_FF) {
			return false
		}

		if (y < 0 || y > CHUNK_SIZE_FF) {
			return false
		}

		return !seen[x or (y shl CHUNK_SHIFT)] && tiles[x or (y shl CHUNK_SHIFT)] != null
	}

	private var widthPositive = 0
	private var widthNegative = 0
	private var heightPositive = 0
	private var heightNegative = 0

	private fun checkLeft(): Boolean {
		for (i in -heightNegative .. heightPositive) {
			if (!filled(rootx - widthNegative, rooty + i)) {
				return false
			}
		}

		return true
	}

	private fun checkRight(): Boolean {
		for (i in -heightNegative .. heightPositive) {
			if (!filled(rootx + widthPositive, rooty + i)) {
				return false
			}
		}

		return true
	}

	private fun checkUp(): Boolean {
		for (i in -widthNegative .. widthPositive) {
			if (!filled(rootx + i, rooty + heightPositive)) {
				return false
			}
		}

		return true
	}

	private fun checkDown(): Boolean {
		for (i in -widthNegative .. widthPositive) {
			if (!filled(rootx + i, rooty - heightNegative)) {
				return false
			}
		}

		return true
	}

	init {
		var expanded = true
		var hitLeft = false
		var hitRight = false
		var hitUp = false
		var hitDown = false

		while (expanded) {
			expanded = false

			// expand left
			if (!hitLeft) {
				widthNegative++

				if (!checkLeft()) {
					widthNegative--
					hitLeft = true
				} else {
					expanded = true
				}
			}

			// expand up
			if (!hitUp) {
				heightPositive++

				if (!checkUp()) {
					heightPositive--
					hitUp = true
				} else {
					expanded = true
				}
			}

			// expand right
			if (!hitRight) {
				widthPositive++

				if (!checkRight()) {
					widthPositive--
					hitRight = true
				} else {
					expanded = true
				}
			}

			// expand down
			if (!hitDown) {
				heightNegative++

				if (!checkDown()) {
					heightNegative--
					hitDown = true
				} else {
					expanded = true
				}
			}
		}

		mins = Vector2i(rootx - widthNegative, rooty - heightNegative)
		maxs = Vector2i(rootx + widthPositive, rooty + heightPositive)
	}

	fun markSeen() {
		for (x in mins.x .. maxs.x) {
			for (y in mins.y .. maxs.y) {
				seen[x or (y shl CHUNK_SHIFT)] = true
			}
		}
	}
}

private fun ccwSortScore(point: Vector2d, axis: Vector2d): Double {
	if (point.x > 0.0) {
		return point.dot(axis)
	} else {
		return -2 - point.dot(axis)
	}
}

/**
 * Чанк мира
 *
 * Хранит в себе тайлы и ентити внутри себя
 *
 * Считается, что один тайл имеет форму квадрата и сторона квадрата примерно равна полуметру,
 * что будет называться Starbound Unit
 *
 * Весь игровой мир будет измеряться в Starbound Unit'ах
 */
abstract class Chunk<WorldType : World<WorldType, This>, This : Chunk<WorldType, This>>(val world: WorldType, val pos: ChunkPos) {
	/**
	 * Возвращает счётчик изменений чанка
	 */
	var changeset = 0
		private set

	fun incChangeset() {
		changeset++
	}

	val aabb = aabbBase + Vector2d(pos.x * CHUNK_SIZE.toDouble(), pos.y * CHUNK_SIZE.toDouble())

	var isPhysicsDirty = false

	fun markPhysicsDirty() {
		if (isPhysicsDirty)
			return

		isPhysicsDirty = true
		world.dirtyPhysicsChunks.add(this as This)
	}

	fun bakeCollisions() {
		foreground.bakeCollisions()
	}

	inner class TileLayer : IMutableTileChunk {
		/**
		 * Возвращает счётчик изменений этого слоя
		 */
		var changeset = 0
			private set

		fun incChangeset() {
			changeset++
			this@Chunk.changeset++
		}

		private val collisionCache = ArrayList<AABB>()
		private val collisionCacheView = Collections.unmodifiableCollection(collisionCache)
		private var collisionChangeset = -1

		private val body = world.physics.createBody(BodyDef(
			position = pos.firstBlock.toDoubleVector(),
			userData = this
		))

		private val collisionChains = ArrayList<B2Fixture>()

		fun bakeCollisions() {
			if (collisionChangeset == changeset)
				return

			collisionChangeset = changeset
			collisionCache.clear()

			for (box in collisionChains) {
				body.destroyFixture(box)
			}

			collisionChains.clear()

			val xyAdd = Vector2d(pos.x * CHUNK_SIZEd, pos.y * CHUNK_SIZEd)
			val seen = BooleanArray(CHUNK_SIZE * CHUNK_SIZE)

			for (y in 0 .. CHUNK_SIZE_FF) {
				for (x in 0..CHUNK_SIZE_FF) {
					if (!seen[x or (y shl CHUNK_SHIFT)] && tiles[x or (y shl CHUNK_SHIFT)] != null) {
						val depthFirst = RectTileFlooderDepthFirst(tiles, seen, x, y)
						val sizeFirst = RectTileFlooderSizeFirst(tiles, seen, x, y)
						val xSpanDepth = depthFirst.maxs.x - depthFirst.mins.x
						val ySpanDepth = depthFirst.maxs.y - depthFirst.mins.y
						val xSpanSize = sizeFirst.maxs.x - sizeFirst.mins.x
						val ySpanSize = sizeFirst.maxs.y - sizeFirst.mins.y

						val aabb: AABB

						if (xSpanDepth * ySpanDepth > xSpanSize * ySpanSize) {
							depthFirst.markSeen()
							aabb = AABB(depthFirst.mins.toDoubleVector(), depthFirst.maxs.toDoubleVector() + Vector2d.POSITIVE_XY)
						} else {
							sizeFirst.markSeen()
							aabb = AABB(sizeFirst.mins.toDoubleVector(), sizeFirst.maxs.toDoubleVector() + Vector2d.POSITIVE_XY)
						}

						collisionChains.add(body.createFixture(FixtureDef(
							shape = PolygonShape().also { it.setAsBox(aabb.width / 2.0, aabb.height / 2.0, aabb.centre, 0.0) },
							friction = 0.4,
						)))

						collisionCache.add(aabb + xyAdd)
					}
				}
			}

			/*
			val seen = BooleanArray(CHUNK_SIZE * CHUNK_SIZE)

			for (y in 0 .. CHUNK_SIZE_FF) {
				for (x in 0 .. CHUNK_SIZE_FF) {
					if (!seen[x or (y shl CHUNK_SHIFT)] && tiles[x or (y shl CHUNK_SHIFT)] != null) {
						val exposed = TileFlooder(tiles, seen, x, y).exposed

						if (exposed.size > 0) {
							val vertices = ArrayList<Vector2d>()
							var centroid = Vector2d.ZERO

							for (pos in exposed) {
								centroid += pos.pos
							}

							centroid /= exposed.size.toDouble()

							for (pos in exposed) {
								if (pos.left) {
									vertices.addU(pos.pos + Vector2d(-0.5, -0.5))
									vertices.addU(pos.pos + Vector2d(-0.5, 0.5))
								}

								if (pos.up) {
									vertices.addU(pos.pos + Vector2d(-0.5, 0.5))
									vertices.addU(pos.pos + Vector2d(0.5, 0.5))
								}

								if (pos.right) {
									vertices.addU(pos.pos + Vector2d(0.5, 0.5))
									vertices.addU(pos.pos + Vector2d(0.5, -0.5))
								}

								if (pos.down) {
									vertices.addU(pos.pos + Vector2d(0.5, -0.5))
									vertices.addU(pos.pos + Vector2d(-0.5, -0.5))
								}
							}

							for (i in vertices.indices) {
								vertices[i] -= centroid
							}

							vertices.sortWith { a, b ->
								val scoreA = ccwSortScore(a.normalized, Vector2d.POSITIVE_Y)
								val scoreB = ccwSortScore(b.normalized, Vector2d.POSITIVE_Y)

								if (scoreA == scoreB) {
									return@sortWith 0
								} else if (scoreA > scoreB) {
									return@sortWith 1
								} else {
									return@sortWith -1
								}
							}

							for (i in vertices.indices) {
								vertices[i] += centroid
							}

							val deflated = ArrayList<Vector2d>()
							var edge = vertices[0]
							deflated.add(edge)

							for (i in 1 until vertices.size) {
								if (vertices[i].x == edge.x) {
									// expand along X axis
									continue
								} else if (vertices[i].y == edge.y) {
									// expand along Y axis
									continue
								}

								// completely different direction

								if (deflated.indexOf(vertices[i - 1]) == -1)
									deflated.add(vertices[i - 1])

								if (deflated.indexOf(vertices[i]) == -1)
									deflated.add(vertices[i])

								edge = vertices[i]
							}

							collisionChains.add(body.createFixture(FixtureDef(
								shape = ChainShape().also { it.createLoop(vertices) },
								friction = 0.4,
							)))
						}
					}
				}
			}*/
		}

		/**
		 * Возвращает список AABB тайлов этого слоя
		 *
		 * Данный список напрямую указывает на внутреннее состояние и будет изменён при перестройке
		 * коллизии чанка, поэтому если необходим стабильный список, его необходимо скопировать
		 */
		fun collisionLayers(): Collection<AABB> {
			if (collisionChangeset != changeset) {
				bakeCollisions()
			}

			return collisionCacheView
		}

		override val pos: ChunkPos
			get() = this@Chunk.pos

		/**
		 * Хранит тайлы как x + y * CHUNK_SIZE
		 */
		private val tiles = arrayOfNulls<ChunkTile>(CHUNK_SIZE * CHUNK_SIZE)

		override operator fun get(x: Int, y: Int): ChunkTile? {
			if (isOutside(x, y))
				return null

			return tiles[x or (y shl CHUNK_SHIFT)]
		}

		operator fun set(x: Int, y: Int, tile: ChunkTile?) {
			if (isOutside(x, y))
				throw IndexOutOfBoundsException("Trying to set tile ${tile?.def?.materialName} at $x $y, but that is outside of chunk's range")

			changeset++
			tiles[x or (y shl CHUNK_SHIFT)] = tile
			markPhysicsDirty()
		}

		override operator fun set(x: Int, y: Int, tile: TileDefinition?): ChunkTile? {
			if (isOutside(x, y))
				throw IndexOutOfBoundsException("Trying to set tile ${tile?.materialName} at $x $y, but that is outside of chunk's range")

			val chunkTile = if (tile != null) ChunkTile(this, tile) else null
			this[x, y] = chunkTile
			changeset++
			markPhysicsDirty()
			return chunkTile
		}

		override fun randomLongFor(x: Int, y: Int): Long {
			return super.randomLongFor(x, y) xor world.seed
		}
	}

	val foreground = TileLayer()
	val background = TileLayer()
	protected val entities = HashSet<Entity>()
	val entitiesAccess = Collections.unmodifiableSet(entities)

	protected abstract fun onEntityAdded(entity: Entity)
	protected abstract fun onEntityTransferedToThis(entity: Entity, otherChunk: This)
	protected abstract fun onEntityTransferedFromThis(entity: Entity, otherChunk: This)
	protected abstract fun onEntityRemoved(entity: Entity)

	fun addEntity(entity: Entity) {
		if (!entities.add(entity)) {
			throw IllegalArgumentException("Already having having entity $entity")
		}

		onEntityAdded(entity)
	}

	fun transferEntity(entity: Entity, otherChunk: Chunk<*, *>) {
		if (otherChunk == this)
			throw IllegalArgumentException("what?")

		if (this::class.java != otherChunk::class.java) {
			throw IllegalArgumentException("Incompatible types: $this !is $otherChunk")
		}

		if (!entities.add(entity)) {
			throw IllegalArgumentException("Already containing $entity")
		}

		onEntityTransferedToThis(entity, otherChunk as This)
		otherChunk.onEntityTransferedFromThis(entity, this as This)

		if (!otherChunk.entities.remove(entity)) {
			throw IllegalStateException("Unable to remove $entity from $otherChunk after transfer")
		}
	}

	fun removeEntity(entity: Entity) {
		if (!entities.remove(entity)) {
			throw IllegalArgumentException("Already not having entity $entity")
		}

		onEntityRemoved(entity)
	}

	override fun toString(): String {
		return "Chunk(pos=$pos, entityCount=${entities.size}, world=$world)"
	}

	companion object {
		val EMPTY = object : IMutableTileChunk {
			override val pos = ChunkPos(0, 0)

			override fun get(x: Int, y: Int): ChunkTile? = null
			override fun set(x: Int, y: Int, tile: TileDefinition?): ChunkTile? = null
		}

		private val aabbBase = AABB(
			Vector2d.ZERO,
			Vector2d(CHUNK_SIZE.toDouble(), CHUNK_SIZE.toDouble()),
		)
	}
}

private fun <E> AbstractList<E>.addU(e: E) {
	if (indexOf(e) == -1) {
		add(e)
	}
}
